import requests
import json
import logging


def call_azure_endpoint(api_url, api_key, request, max_timeouts=3, timeout_time=1.2):
    """
    Wrapper to call an Azure Machine Learning endpoint which also handles timeouts and errors.
    """
    response = {
        'content': None,
        'success': False,
        'original_response': None,
    }
    request_data = json.dumps(request)
    # Set the content type
    headers = {'Content-Type': 'application/json'}
    # If authentication is enabled, set the authorization header
    headers['Authorization'] = f'Bearer {api_key}'
    resp = None
    while (resp is None or resp.text is None) and max_timeouts >= 0:
        try:
            resp = requests.post(api_url, request_data, headers=headers, timeout=timeout_time)
        except requests.exceptions.Timeout:
            logging.warning(f'Request to {api_url} timed out (TimeOut)...')
        except requests.exceptions.ConnectionError:
            logging.warning(f'Request to {api_url} timed out (ConnectionError)...')
        max_timeouts -= 1

    if resp is None or resp.text is None or resp.status_code >= 400:
        response['original_response'] = resp
        return response
    else:
        response = {
            'content': json.loads(resp.text),
            'success': True,
            'original': resp,
        }
        return response




import threading
import asyncio
import json
from uuid import uuid4
from datetime import datetime
import logging
from functools import partial

import tornado.ioloop
import tornado.web
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.json_format import MessageToDict

from delphai_backend_utils.config import get_config
from delphai_backend_utils.grpc_calls import _message_template

THIS_UUID = str(uuid4())
STARTED_TIME = datetime.now().strftime('%Y-%m-%d %H:%M:%S')


class RequestHandler(tornado.web.RequestHandler):

    async def get(self, path):
        method_descriptors = self.settings["method_descriptors"]
        grpc_host_and_port = get_config('server.host_and_port') or '0.0.0.0:8080'
        if not path:
            # Respond a list of functions
            res = {
                'gateway_id': THIS_UUID,
                'started': STARTED_TIME,
                'status': 'active',
                'functions': [{
                    'function_name': meth_name,
                    'branches': [{'name': 'not-defined', 'host': grpc_host_and_port, 'is_default': True}]
                } for meth_name in method_descriptors.keys()]
            }
            self.write(res)
        elif path not in self.settings["method_descriptors"]:
            self._reason = 'Wrong function name'
            self._status_code = 404
            self.finish(f'Wrong function name. Available functions: {", ".join(method_descriptors.keys())}')
            return
        else:
            method_descriptor = self.settings["method_descriptors"][path]
            self.write({
                'function_name': path,
                'branch': 'not-defined',
                'host': grpc_host_and_port,
                'input': _message_template(method_descriptor['input_def']),
                'output': _message_template(method_descriptor['output_def']),
            })
        self.finish()

    async def post(self, path):
        # service_descriptor = self.settings["service_descriptor"]
        if not path:
            self._reason = 'Function name not specified'
            self._status_code = 400
            self.finish('Full function name should be specified in path')
            return
        method_descriptors = self.settings["method_descriptors"]
        if path not in self.settings["method_descriptors"]:
            self._reason = 'Wrong function name'
            self._status_code = 404
            self.finish(f'Wrong function name. Available functions: {", ".join(method_descriptors.keys())}')
            return
        if self.request.body:
            try:
                in_msg_dict = json.loads(self.request.body)
            except Exception as exc:
                self._reason = 'Wrong JSON'
                self._status_code = 400
                self.finish(f'Wrong JSON: {exc}')
                return
        else:
            in_msg_dict = {}
        if not isinstance(in_msg_dict, dict):
            self._reason = 'Wrong JSON'
            self._status_code = 400
            self.finish('Wrong JSON: must be a single object')
            return

        try:
            in_msg = method_descriptors[path]['input_type'](**in_msg_dict)
        except Exception as exc:
            self._reason = 'Wrong message format'
            self._status_code = 400
            self.finish(f'Wrong message format: {exc}')
            return

        try:
            loop = asyncio.get_event_loop()
            func = partial(method_descriptors[path]['method'], in_msg, None)  # TODO: implement context parameter
            res = await loop.run_in_executor(None, func)
        except Exception as exc:
            self._reason = 'Internal Server Error'
            self._status_code = 500
            self.finish(f'Error: {exc}')
            return

        self.write(MessageToDict(res, preserving_proto_field_name=True))
        self.finish()


class WebServer(threading.Thread):
    service_descriptor = None
    grpc_server = None

    def run(self):
        gateway_host_and_port = get_config('server.gateway_host_and_port') or '0.0.0.0:7070'
        split_host_and_port = gateway_host_and_port.split(':')
        if len(split_host_and_port) == 1:
            raise ValueError('Wrong server.gateway_host_and_port value. Expected: <host>:<port>')
        if not split_host_and_port[-1].isnumeric():
            raise ValueError('Wrong server.gateway_host_and_port value. Port must be numeric')

        _sym_db = _symbol_database.Default()
        method_descriptors = {}  # full name -> {method, input_type, output_type}
        for service_handler in self.grpc_server._state.generic_handlers:
            if service_handler._name.startswith('grpc.reflection.'):
                continue
            for meth_key, meth_handler in service_handler._method_handlers.items():
                meth_key = meth_key[1:].replace('/', '.')
                service_name, method_name = meth_key.split('.')[2:]
                meth_descr = self.service_descriptor.services_by_name[service_name].methods_by_name[method_name]
                method_descriptors[meth_key] = {
                    'method': meth_handler.unary_unary,
                    'input_def': meth_descr.input_type,
                    'output_def': meth_descr.output_type,
                    'input_type': _sym_db.GetPrototype(meth_descr.input_type),
                    'output_type': _sym_db.GetPrototype(meth_descr.output_type),
                }

        asyncio.set_event_loop(asyncio.new_event_loop())
        application = tornado.web.Application([
            (r"/(.*)", RequestHandler)],
            default_host=':'.join(split_host_and_port[:-1]),
            service_descriptor=self.service_descriptor,
            method_descriptors=method_descriptors)
        application.listen(int(split_host_and_port[-1]))
        logging.info(f'Started gateway {gateway_host_and_port}')
        tornado.ioloop.IOLoop.instance().start()


def run_own_gateway(service_descriptor, grpc_server):
    server = WebServer(daemon=True)
    server.service_descriptor = service_descriptor
    server.grpc_server = grpc_server
    server.start()



from base64 import b64decode
import json


def get_user(context):
    metadata = dict(context.invocation_metadata())
    user64 = metadata['x-delphai-user']
    user = b64decode(user64)
    return json.loads(user)




from logging import error, warn, info, debug  # noqa: F401
from sys import exc_info
from traceback import format_tb

import coloredlogs
from logging.config import dictConfig
from delphai_backend_utils.config import get_config


def _configure_logging():
    logging_config = get_config('logging')
    standard_format = logging_config['formatters']['standard']
    dictConfig(logging_config)
    coloredlogs.install(fmt=standard_format['format'], datefmt=standard_format['datefmt'])


_configure_logging()


def error_with_traceback():
    """
    Writes to log an occured exception with a last line of traceback
    """
    einfo = exc_info()
    exc_class = einfo[0].__name__
    exc_descr = str(einfo[1])
    traceback = format_tb(einfo[2])
    traceback = '\n' + traceback[-1].strip() if traceback else ''
    error(f'{exc_class}: {exc_descr}{traceback}')



"""
Access to other grpc microservices. Functions:
- call_grpc - calls other grpc service
- async_call_grpc - awaitable version of call_grpc. Use if caller is "async def".
- get_description - to get other grpc's service description
- async_get_description - awaitable version of get_description. Use if caller is "async def".

See "Public API" section below.
"""

import asyncio
from base64 import b64encode
from collections import defaultdict
from functools import partial
import json
import logging
from time import perf_counter, sleep
from uuid import uuid4
import re

import grpc
from grpc_reflection.v1alpha import reflection_pb2, reflection_pb2_grpc
from google.protobuf import descriptor_pb2
from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.symbol_database import SymbolDatabase
from google.protobuf import json_format
from kubernetes import client as k8s_client, config as k8s_config
from kubernetes.client.api_client import ApiClient as K8sApiClient
import requests
from memoization import cached

from delphai_backend_utils.caching import Cacher, register_cacher, get_cache
from delphai_backend_utils.config import get_config


GET_SERVICES_REQUEST = reflection_pb2.ServerReflectionRequest(list_services='services')
EXCUDED_NAMESPACES = set([
    'kube-public', 'kube-system', 'elasticsearch', 'gateway',
    'lens-metrics', 'nginx-ingress', 'default', 'kube-node-lease'])
CACHER_NAME = 'external services'
THIS_UUID = str(uuid4())  # Used to prevent self-gatewaying


class ExternalServices(Cacher):
    """
    Explored external services cache
    """
    name = CACHER_NAME
    renew_interval = 300  # renew every 5 min
    timeout = 60
    incluster = None  # initially

    def get_data_func(self):
        """Returns dict: function full name -> dict:
        - None: default branch definition
        - <branch name>: branch definition
        where branch definition is a dict:
        - host (str)
        - is_default (bool)
        - For functions that called in "natural" way
          - service_name (str)
          - method_name (str)
          - method_descriptor
          - input_descriptor
          - output_descriptor
          - input_prototype
          - output_prototype
        - For functions that called through gateway:
          - gateway: address of gateway
        """
        sleep(1)  # Sleep a second to allow this server to start

        strt = perf_counter()
        res = defaultdict(dict)
        is_first_update = not self.is_ready()

        if self.incluster is None:
            self.incluster = False
            try:
                k8s_config.load_incluster_config()
                self.incluster = True
                logging.info('Connected to kubernetes in cluster')
            except Exception:
                logging.warn('Working outside a kubernetes cluster')
                # k8s_config.load_kube_config()
                # with K8sApiClient() as api:
                #     logging.info(f'Connected to external kubernetes: {api.configuration.host}')

        def _clean_addr(addr: str) -> str:
            if not addr:
                return None
            while addr.endswith('/'):
                addr = addr[:-1]
            return addr or None

        grpc_config = get_config('grpc') or {}
        cfg_verbose = bool(grpc_config.get('verbose'))
        cfg_explore_localhost = bool(grpc_config.get('explore_localhost'))
        cfg_default_gateway = _clean_addr(grpc_config.get('default_gateway'))
        cfg_function_branches = {}
        for func_name, func_inst in (grpc_config.get('function_branches') or {}).items():
            if func_name and func_inst:
                clean_func_name = func_name.replace('/', '.')
                while clean_func_name.endswith('.*'):
                    clean_func_name = clean_func_name[:-2]
                if '*' in clean_func_name:
                    if is_first_update:
                        logging.warn(f'Bad config key "grpc.function_branches.{func_name}". Wrong position of "*".')
                    continue
                cfg_function_branches[clean_func_name + '.'] = func_inst

        gw_data = None
        if cfg_default_gateway:
            # Explore gatewayed functions
            gw_err = None
            try:
                gw_data_resp = requests.get(cfg_default_gateway)
                if gw_data_resp.status_code < 400:
                    gw_data = gw_data_resp.json()
                else:
                    gw_err = f'Status code from default gateway {cfg_default_gateway}: {gw_data_resp.status_code}'
            except Exception as exc:
                gw_err = f'Failed to use default gateway {cfg_default_gateway}: {exc}'
            if gw_data:
                if 'gateway_id' not in gw_data or 'status' not in gw_data or 'functions' not in gw_data:
                    gw_err = f'Wrong response from the default gateway {cfg_default_gateway}'
                    gw_data = None
                elif gw_data['gateway_id'] == THIS_UUID:
                    gw_err = f'Will not use myself as a default gateway {cfg_default_gateway}'
                    gw_data = None
                elif gw_data['status'] != 'active':
                    gw_err = f'Status of the default gateway {cfg_default_gateway} is "{gw_data["status"]}"'
                    gw_data = None
                else:
                    if cfg_verbose and is_first_update:
                        logging.info(f'Will use {cfg_default_gateway} as a default gateway')
            if gw_err:
                if is_first_update:
                    logging.warn(gw_err)

        # >>>> nested func connect_n_explore start
        def _connect_n_explore(host: str, branch_name: str, namespace_name: str):
            try:
                # Request reflection info
                with grpc.insecure_channel(host) as channel:
                    reflection_client = reflection_pb2_grpc.ServerReflectionStub(channel)
                    for response in reflection_client.ServerReflectionInfo(x for x in [GET_SERVICES_REQUEST]):
                        service_names = [service.name for service in response.list_services_response.service]
                        file_requests = (
                            reflection_pb2.ServerReflectionRequest(file_containing_symbol=service_name)
                            for service_name in service_names
                        )
                        file_descriptor_proto_bytes = _get_proto_bytes_for_requests(reflection_client, file_requests)

                        descriptor_pool = DescriptorPool()
                        _add_protos_to_descriptor_pool(reflection_client, descriptor_pool, file_descriptor_proto_bytes)
                        symbol_database = SymbolDatabase(descriptor_pool)

                        for sname in service_names:
                            sdescr = descriptor_pool.FindServiceByName(sname)
                            for methdescr in sdescr.methods:
                                meth_full_name = methdescr.full_name
                                if meth_full_name.startswith('grpc.reflection.v1alpha.ServerReflection'):
                                    continue
                                if branch_name not in res[meth_full_name]:
                                    inst_definition = {
                                        'host': host,
                                        'branch': branch_name,
                                        'is_default': False,
                                        'service_name': sname,
                                        'method_name': methdescr.name,
                                        'method_descriptor': methdescr,
                                        'input_descriptor': methdescr.input_type,
                                        'output_descriptor': methdescr.output_type,
                                        'input_prototype': symbol_database.GetPrototype(methdescr.input_type),
                                        'output_prototype': symbol_database.GetPrototype(methdescr.output_type),
                                    }
                                    res[meth_full_name][branch_name] = inst_definition
            except Exception:  # as exc:
                pass
        # <<<< nested func connect_n_explore end

        # Explore addresses specified in grpc.function_branches
        fpatterns_by_addr = defaultdict(dict)
        localhost_finder = re.compile('^https?://([^:/]+?):([0-9]+)$')
        if not self.incluster and cfg_function_branches:
            for func_pattern, br_str in cfg_function_branches.items():
                lh_match = localhost_finder.match(br_str)
                if lh_match:
                    host, port = lh_match.groups()
                    fpatterns_by_addr[(host, port)][func_pattern] = 1
            for host_port in fpatterns_by_addr.keys():
                if cfg_verbose and is_first_update:
                    logging.info(f'Exploring gRPC services on local port {host_port[1]}...')
                svc_local_addr = f'{host_port[0]}:{host_port[1]}'
                _connect_n_explore(svc_local_addr, f'{svc_local_addr}', svc_local_addr)

        # Connect to Kubernetes and explore registered services.
        if self.incluster:
            with K8sApiClient() as api:
                v1 = k8s_client.CoreV1Api(api)
                namespaces = v1.list_namespace()
                for namespace in filter(lambda ns: ns.metadata.name not in EXCUDED_NAMESPACES, namespaces.items):
                    for service in v1.list_namespaced_service(namespace.metadata.name).items:
                        svc_md = service.metadata
                        if svc_md is not None and service.spec is not None \
                           and service.spec.ports:
                            all_ports = set(str(el.target_port) for el in service.spec.ports if el is not None)
                            if '8080' in all_ports:
                                svc_local_addr = f'{svc_md.name}.{svc_md.namespace}:8080'
                                _connect_n_explore(svc_local_addr, svc_md.name, svc_md.namespace)
        elif cfg_explore_localhost:
            for local_port in range(8080, 8089):  # Explore local ports 8080..8088
                if ('localhost', str(local_port)) not in fpatterns_by_addr:
                    if cfg_verbose and is_first_update:
                        logging.info(f'Exploring gRPC services on local port {local_port}...')
                    svc_local_addr = f'localhost:{local_port}'
                    _connect_n_explore(svc_local_addr, f'{svc_local_addr}', svc_local_addr)

        if gw_data:
            for func_gw_descr in gw_data['functions']:
                func_gw_name = func_gw_descr['function_name']
                for inst_gw_descr in func_gw_descr['branches']:
                    branch_name = inst_gw_descr['name']
                    if branch_name not in res[func_gw_name]:
                        res[func_gw_name][branch_name] = {
                            'host': f'{cfg_default_gateway} -> {inst_gw_descr["host"]}',
                            'branch': branch_name,
                            'is_default': False,
                            'gateway': cfg_default_gateway,
                        }

        # Set default definitions
        reported_wrongs = set()
        for meth_name, definitions in res.items():
            # Apply priority specified in grpc.function_branches
            for func_pattern, br_str in cfg_function_branches.items():
                if not (meth_name + '.').startswith(func_pattern):
                    continue
                lh_match = localhost_finder.match(br_str)
                if lh_match:
                    host, port = lh_match.groups()
                    svc_local_addr = f'{host}:{port}'
                    if svc_local_addr in definitions:
                        definitions[None] = definitions[svc_local_addr]
                        break
                    elif is_first_update:
                        logging.warn(f'Function "{meth_name}" not found on {br_str}')
                else:
                    if br_str.startswith('branch://'):
                        branch = br_str[9:]
                        if branch in definitions:
                            definitions[None] = definitions[branch]
                            break
                        elif branch.startswith('svc-') and branch[4:] in definitions:
                            definitions[None] = definitions[branch[4:]]
                            break
                        elif not branch.startswith('svc-') and f'svc-{branch}' in definitions:
                            definitions[None] = definitions[f'svc-{branch}']
                            break
                        elif is_first_update and br_str not in reported_wrongs:
                            logging.warn(f'Branch "{branch}" not found')
                            reported_wrongs.add(br_str)
                    else:
                        if is_first_update and br_str not in reported_wrongs:
                            logging.warn(
                                f'Wrong branch definition "{br_str}" in grpc.function_branches. '
                                f'Use "branch://" or "http://" prefixes.')
                            reported_wrongs.add(br_str)
            if None not in definitions:
                if 'svc-master' in definitions:
                    definitions[None] = definitions['svc-master']
                else:
                    definitions[None] = definitions[list(definitions.keys())[0]]
            definitions[None]['is_default'] = True

        if cfg_verbose:
            if is_first_update:
                if res:
                    ffuncs = []
                    for fidx, ffullname in enumerate(sorted(res.keys())):
                        ffuncs.append(
                            f'{fidx+1}. {ffullname}:\n' + '\n'.join(
                                f' {"*" if bv["is_default"] else " "} branch {bk} at {bv["host"]}'
                                for bk, bv in res[ffullname].items() if bk))
                    found_msg = '\n' + '\n'.join(ffuncs)
                    logging.info(f'Functions found: {found_msg}')
                else:
                    logging.warn('NO functions found')
            logging.info(f'{self.name} cache was updated in {perf_counter()-strt:.3f}s.')
        return res


register_cacher(ExternalServices, True)


def _get_proto_bytes_for_requests(reflection_client, file_requests):
    """
    Borrowed from eagr.reflection.reflection_descriptor_database.get_proto_bytes_for_requests
    """
    file_descriptors_responses = reflection_client.ServerReflectionInfo(file_requests)
    proto_bytes = [
        file_descriptor_proto
        for response in file_descriptors_responses
        for file_descriptor_proto in response.file_descriptor_response.file_descriptor_proto
    ]
    return proto_bytes


def _add_protos_to_descriptor_pool(reflection_client, descriptor_pool, file_descriptor_proto_bytes):
    """
    Borrowed from eagr.reflection.reflection_descriptor_database.add_protos_to_descriptor_pool
    """
    # Set of already imported names
    imported_names = set()
    # Map of file descriptor proto names to the respective file descriptor protos
    reflected_names = {}
    # Stack of file descriptor proto names to be imported
    proto_names = []

    # Load the initial proto name stack and reflected names hash with root-level protos to import
    for serialized_proto in file_descriptor_proto_bytes:
        parsed_file_descriptor_proto = descriptor_pb2.FileDescriptorProto()
        parsed_file_descriptor_proto.ParseFromString(serialized_proto)
        file_descriptor_proto_name = parsed_file_descriptor_proto.name  # pylint: disable=no-member
        proto_names.append(file_descriptor_proto_name)
        reflected_names[file_descriptor_proto_name] = parsed_file_descriptor_proto

    # Recursively find file descriptor proto dependencies, and import the entire tree bottom-up
    for proto_name in proto_names:
        _import_dependencies_then_proto(
            reflection_client, descriptor_pool, proto_name, imported_names, reflected_names
        )


def _import_dependencies_then_proto(
    reflection_client, descriptor_pool, proto_name, imported_names, reflected_names
):
    """
    Borrowed from eagr.reflection.reflection_descriptor_database.import_dependencies_then_proto
    """
    if proto_name in imported_names:
        return

    # Anything in the proto_name depth-first search stack should already be reflected
    if proto_name not in reflected_names:
        raise RuntimeError(
            "Something went wrong. Stacked planned imports should either "
            "be already reflected or already imported. Please fix."
        )

    dependencies = reflected_names[proto_name].dependency  # pylint: disable=no-member
    # Reflect not-yet-reflected dependencies the same way we run server reflection on root protos
    not_yet_reflected_dependencies = []
    for dependency in dependencies:
        if (dependency not in imported_names) and (dependency not in reflected_names):
            not_yet_reflected_dependencies.append(dependency)

    if not_yet_reflected_dependencies:
        file_requests = (
            reflection_pb2.ServerReflectionRequest(file_by_filename=dependency)
            for dependency in not_yet_reflected_dependencies
        )
        file_descriptor_proto_bytes = _get_proto_bytes_for_requests(reflection_client, file_requests)
        for serialized_proto in file_descriptor_proto_bytes:
            parsed_file_descriptor_proto = descriptor_pb2.FileDescriptorProto()
            parsed_file_descriptor_proto.ParseFromString(serialized_proto)
            file_descriptor_proto_name = (
                parsed_file_descriptor_proto.name  # pylint: disable=no-member
            )
            reflected_names[file_descriptor_proto_name] = parsed_file_descriptor_proto

    # Recursively import dependencies
    for dependency in dependencies:
        _import_dependencies_then_proto(
            reflection_client, descriptor_pool, dependency, imported_names, reflected_names
        )

    # Import the proto itself, update memos
    descriptor_pool.Add(reflected_names[proto_name])
    imported_names.add(proto_name)
    del reflected_names[proto_name]


def _find_meth_definition(function: str, branch: str) -> dict:
    branch = branch or None
    cacher = get_cache(ExternalServices.name)
    methods_dict = cacher.get_data()
    if function not in methods_dict:
        raise ServiceFunctionNotFound(f'Service function "{function}" not found')
    meth_definitions = methods_dict[function]

    if branch:
        if branch not in meth_definitions:
            branch_found = False
            if branch.startswith('svc-'):
                # try to find without "svc-"
                if branch[4:] in meth_definitions:
                    branch = branch[4:]
                    branch_found = True
            else:
                # try to find with "svc-"
                if f'svc-{branch}' in meth_definitions:
                    branch = f'svc-{branch}'
                    branch_found = True
            if not branch_found:
                raise ServiceNotFound(f'Branch "{branch}" of for function "{function}" not found')
    else:
        branch = None
    return meth_definitions[branch]


def _message_template(message_descriptor, lvls: int = 10):
    """
    Returns descriptive template for message.
    """
    if lvls == 0:
        return '...'
    res = {}
    for fld in message_descriptor.fields:
        if fld.message_type:
            fld_content = _message_template(fld.message_type, lvls - 1)
        else:
            fld_content = fld.default_value
        if fld.default_value == []:
            if fld.message_type:
                fld_content = [fld_content, ]
            elif fld.type == 9:  # string
                fld_content = ['', ]
            elif fld.type == 2:  # float
                fld_content = [0., ]
            elif fld.type == 8:  # boolean
                fld_content = [False, ]
            elif fld.type in [3, 5]:  # int32, int64
                fld_content = [0, ]
        res[fld.name] = fld_content
    return res


# ====================== Public API ======================


class GrpcException(Exception):
    """General gRPC exception"""


class ServiceNotFound(GrpcException):
    """Service not found"""


class ServiceFunctionNotFound(GrpcException):
    """Service function not found"""


def call_grpc(function: str, params: dict, metadata: dict = None, branch: str = None) -> dict:
    """
    Calls other grpc service.
    :param function: function name, e.g. "delphai.textgen.Textgen.macro"
    :param params: request data as dict
    :param metadata: metadata as a dict to pass to service
    :param branch: specify branch name if you need particular service instance
    :return: response object as dict
    """
    meth_definition = _find_meth_definition(function, branch)

    res = None
    if 'gateway' in meth_definition:
        try:
            gw_addr = f'{meth_definition["gateway"]}/{function}'
            if branch:
                gw_addr = f'{gw_addr}?branch={branch}'
            headers = {}
            if metadata:
                headers['gRPC-Metadata'] = b64encode(json.dumps(metadata).encode())
            gw_data_resp = requests.post(gw_addr, json=params, headers=headers)
            if gw_data_resp.status_code < 400:
                res = gw_data_resp.json()
            else:
                raise GrpcException(
                    f'Status code from the gateway {meth_definition["gateway"]}: {gw_data_resp.status_code}')
        except Exception as exc:
            raise GrpcException(f'Failed to call through the gateway {meth_definition["gateway"]}: {exc}')
    else:
        with grpc.insecure_channel(meth_definition['host'], (("grpc.lb_policy_name", "round_robin"),)) as channel:
            method_callable = channel.unary_unary(
                "/{}/{}".format(meth_definition['service_name'], meth_definition['method_descriptor'].name),
                request_serializer=meth_definition['input_prototype'].SerializeToString,
                response_deserializer=meth_definition['output_prototype'].FromString,
            )
            res = json_format.MessageToDict(
                method_callable(
                    meth_definition['input_prototype'](**params), metadata=metadata.items()
                    if isinstance(metadata, dict) else None),
                preserving_proto_field_name=True, including_default_value_fields=False)
    if res is None:
        raise GrpcException(f'Function {function} was not called')
    elif len(res) == 1 and 'error' in res:
        raise GrpcException(f'Function {function} call failed: {res["error"]}')
    return res


async def async_call_grpc(function: str, params: dict, metadata: dict = None, branch: str = None) -> dict:
    """
    Awaitable version of call_grpc. Use if caller is "async def".
    """
    loop = asyncio.get_event_loop()
    func = partial(call_grpc, function, params, metadata, branch)
    return await loop.run_in_executor(None, func)


@cached(ttl=30)
def get_description(function: str, branch: str = None) -> dict:
    """
    Get other grpc's service description.
    :param function: function name, e.g. "delphai.textgen.Textgen.macro"
    :param branch: specify branch name if you need particular service instance
    :return: dict(function_name, branch, host, input template, output template)
    """
    meth_definition = _find_meth_definition(function, branch)

    if 'gateway' in meth_definition:
        res = None
        try:
            gw_addr = f'{meth_definition["gateway"]}/{function}'
            if branch:
                gw_addr = f'{gw_addr}?branch={branch}'
            gw_data_resp = requests.get(gw_addr)
            if gw_data_resp.status_code < 400:
                res = gw_data_resp.json()
            else:
                raise GrpcException(
                    f'Status code from the gateway {meth_definition["gateway"]}: {gw_data_resp.status_code}')
        except Exception as exc:
            raise GrpcException(f'Failed to get description through the gateway {meth_definition["gateway"]}: {exc}')
        res['host'] = meth_definition['host']
        return res
    else:
        return {
            'function_name': function,
            'branch': meth_definition['branch'],
            'host': meth_definition['host'],
            'input': _message_template(meth_definition['input_descriptor']),
            'output': _message_template(meth_definition['output_descriptor']),
        }


async def async_get_description(function: str, branch: str = None) -> dict:
    """
    Awaitable version of get_description. Use if caller is "async def".
    """
    loop = asyncio.get_event_loop()
    func = partial(get_description, function, branch)
    return await loop.run_in_executor(None, func)



from base64 import b64decode
import os
import logging
from time import sleep

from omegaconf import OmegaConf
from memoization import cached
from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient
from dotenv import load_dotenv
from kubernetes import client as kube_client, config as kube_config
from kubernetes.config import ConfigException

load_dotenv()


def __connect_to_kubernetes():
    try:
        kube_config.load_incluster_config()
    except Exception as ex:
        logging.debug(str(ex))
        kube_config.load_kube_config()


def __get_kubernetes_secret(path: str):
    try:
        __connect_to_kubernetes()
    except ConfigException:
        logging.warn(f'Failed to obtain a secret value "kube:{path}": no available configuration')
        return '???'
    except Exception as exc:
        logging.warn(f'Failed to obtain a secret value "kube:{path}": {type(exc).__name__}')
        return '???'

    split = path.split('/')
    namespace = 'default'
    secret_name = split[0]
    secret_key = split[1]
    if len(split) > 2:
        namespace = split[0]
        secret_name = split[1]
        secret_key = split[2]
    v1 = kube_client.CoreV1Api()

    try:
        secret = v1.read_namespaced_secret(secret_name, namespace).data[secret_key]
    except Exception as exc:
        logging.warn(f'Failed to obtain a secret value "kube:{path}": {type(exc).__name__}')
        return '???'

    secret = b64decode(secret).decode('utf-8')
    return secret


def __get_secrets_client():
    delphai_environment = os.environ.get('DELPHAI_ENVIRONMENT')
    credential = DefaultAzureCredential()
    return SecretClient(f'https://delphai-{delphai_environment}-kv.vault.azure.net', credential)


def __get_secret(name: str):
    try:
        secrets_client = __get_secrets_client()
        res = secrets_client.get_secret(name).value
    except Exception as exc:
        logging.warn(f'Failed to obtain a secret value "azkv:{name}": {type(exc).__name__}')
        return '???'
    return res


def __get_keyring_secret(name: str):
    import keyring
    service_n_user = name.split('/')
    service_name = '/'.join(service_n_user[:-1])
    username = ''.join(service_n_user[-1:])
    return keyring.get_password(service_name, username) or '???'


OmegaConf.register_resolver('azkv', __get_secret)
OmegaConf.register_resolver('kube', __get_kubernetes_secret)
OmegaConf.register_resolver('keyring', __get_keyring_secret)


CURRENT_CONFIG = None
CONFIG_IS_LOADING = False


def __load_config():
    global CURRENT_CONFIG, CONFIG_IS_LOADING
    while CONFIG_IS_LOADING:
        sleep(0.05)
    if CURRENT_CONFIG is not None:
        return CURRENT_CONFIG
    while CONFIG_IS_LOADING:
        sleep(0.05)
    CONFIG_IS_LOADING = True
    try:
        if 'DELPHAI_ENVIRONMENT' not in os.environ:
            raise Exception('DELPHAI_ENVIRONMENT is not defined')
        default_config = OmegaConf.load('./config/default.yml')
        delphai_environment = os.environ.get('DELPHAI_ENVIRONMENT')
        if os.path.isfile(f'./config/{delphai_environment}.yml'):
            delphai_env_config = OmegaConf.load(f'./config/{delphai_environment}.yml')
        else:
            delphai_env_config = OmegaConf.create()
        CURRENT_CONFIG = OmegaConf.merge(default_config, delphai_env_config)
        OmegaConf.set_readonly(CURRENT_CONFIG, True)
        return CURRENT_CONFIG
    finally:
        CONFIG_IS_LOADING = False


@cached
def get_config(path: str = ''):
    config = __load_config()
    if path is None:
        return config
    selected = OmegaConf.select(config, path)
    if OmegaConf.is_config(selected):
        return OmegaConf.to_container(selected, resolve=True)
    else:
        return selected



"""
Connection to the database
"""

from multiprocessing import current_process

import pymongo

from delphai_backend_utils.config import get_config

DB_CONN_BY_PID_AND_DBNAME = {}  # (process id, db_name) -> DB_CONN

_db_params = {
    'connection_string': get_config('db')['connection_string'],
    'default_db_name': get_config('db')['name'],
}


def get_own_db_connection(db_name: str = None) -> pymongo.database.Database:
    """
    Creates neq connection to database.
    :param db_name: use this parameter only if DB name differs from _db_params["default_db_name"]
    :return: database connection
    """
    pid = current_process().pid
    # db_config = get_config('db')
    req_db_name = _db_params['default_db_name'] if db_name is None else db_name
    if (pid, req_db_name) in DB_CONN_BY_PID_AND_DBNAME:
        return DB_CONN_BY_PID_AND_DBNAME[(pid, req_db_name)]
    client = pymongo.MongoClient(_db_params['connection_string'])

    res = client[req_db_name]
    DB_CONN_BY_PID_AND_DBNAME[(pid, req_db_name)] = res
    return res


def chunks(iterable, chunk_size=100):
    """
    Splits iterable stream by chunks of size chunk_size. Very usefull when we need to split read
    or write operations to butches of reasonable size.
    :param iterable: something interable
    :param chunk_size: desirable size if chunks to be produced
    :yield: lists of elements extracted from iterable
    """
    curr_chunk = []
    for val in iterable:
        if len(curr_chunk) >= chunk_size:
            yield curr_chunk
            curr_chunk = []
        curr_chunk.append(val)
    if curr_chunk:
        yield curr_chunk



from setuptools import setup


def readme():
    with open('README.md') as f:
        README = f.read()
    return README


setup(
    name="delphai_backend_utils",
    version="1.0.8",
    description="A Python package to manage kube secrets.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/delphai/delphai-backend-utils",
    author="Delphai",
    author_email="ahmed.mahmoud@delphai.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["delphai_backend_utils"],
    include_package_data=True,
    install_requires=[
        'omegaconf', 'memoization',
        'azure-identity==1.4.0b3', 'azure-keyvault',
        'python-dotenv', 'coloredlogs',
        'kubernetes', 'keyring'],
)
